from typing import List

from torch.utils.tensorboard import SummaryWriter


class AverageMeter:
    """Compute and store the average and current value"""

    def __init__(self, name, fmt=":f"):
        self.name = name
        self.fmt = fmt
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

    def __str__(self):
        fmtstr = "{name} {val" + self.fmt + "} ({avg" + self.fmt + "})"
        return fmtstr.format(**self.__dict__)


class MetricsTracker:
    def __init__(self, metrics: List[str]):
        self.metrics = metrics
        self.trackers = {m: AverageMeter(m) for m in self.metrics}
        if "loss" not in self.trackers:
            self.trackers["loss"] = AverageMeter("loss")

    def update(self, results):
        for metric_name, tracker in self.trackers.items():
            tracker.update(results[metric_name])

    def reset(self):
        for tracker in self.trackers.values():
            tracker.reset()

    def avg(self):
        avg_results = {}
        for metric_name, tracker in self.trackers.items():
            avg_results[metric_name] = tracker.avg

        return avg_results


def write_to_tensorboard(writer: SummaryWriter, logs: dict, mode: str, step: int):
    for metric_name, value in logs.items():
        writer.add_scalar(f"{mode}/{metric_name}", value, step)



from typing import List

import numpy as np
import torch

from hezar.constants import MetricType

from ...configs import TrainerConfig
from ...data.datasets import Dataset
from ...models import Model
from ...utils import get_logger
from ..trainer import Trainer


logger = get_logger(__name__)


class TextClassificationTrainer(Trainer):
    """
    A trainer for all text classification models

    Args:
        model ([`Model`] or `torch.nn.Module`): The main model to train and evaluate
        config (TrainerConfig): Training configuration and parameters
        train_dataset (Dataset): Train dataset
        eval_dataset (Dataset): Evaluation dataset
        data_collator: Collate function, usually included in the dataset object itself
        optimizer (optim.Optimizer): Model optimizer
        lr_scheduler: Optional learning-rate scheduler

    """
    AVAILABLE_METRICS = [
        MetricType.ACCURACY,
        MetricType.RECALL,
        MetricType.PRECISION,
        MetricType.F1,
    ]

    def __init__(
        self,
        model: Model = None,
        config: TrainerConfig = None,
        train_dataset: Dataset = None,
        eval_dataset: Dataset = None,
        data_collator=None,
        optimizer: torch.optim.Optimizer = None,
        lr_scheduler=None,
    ):
        super().__init__(
            model=model,
            config=config,
            train_dataset=train_dataset,
            eval_dataset=eval_dataset,
            data_collator=data_collator,
            optimizer=optimizer,
            lr_scheduler=lr_scheduler,
        )
        self.criterion = torch.nn.CrossEntropyLoss(ignore_index=-100)

    def compute_loss(self, logits, labels, **kwargs) -> torch.Tensor:
        loss = self.criterion(logits, labels)
        return loss

    def compute_metrics(self, predictions: List[np.ndarray], labels: List[np.ndarray], **kwargs):
        predictions = np.array(predictions).argmax(1).flatten()
        labels = np.array(labels).flatten()
        results = {}
        for metric_name, metric in self.metrics.items():
            results[metric_name] = metric.compute(predictions, labels, average="macro")
        return results




"""
A BERT model for sequence labeling built using HuggingFace Transformers
"""
from typing import Dict, List, Union

from torch import nn
from transformers import BertConfig, BertModel

from ....models import Model
from ....registry import register_model
from .bert_sequence_labeling_config import BertSequenceLabelingConfig


@register_model("bert_sequence_labeling", BertSequenceLabelingConfig)
class BertSequenceLabeling(Model):
    def __init__(self, config: BertSequenceLabelingConfig, **kwargs):
        super().__init__(config, **kwargs)
        self.bert = BertModel(self._build_inner_config())
        classifier_dropout = (
            config.classifier_dropout if config.classifier_dropout is not None else config.hidden_dropout_prob
        )
        self.dropout = nn.Dropout(classifier_dropout)
        self.classifier = nn.Linear(config.hidden_size, config.num_labels)

    def _build_inner_config(self):
        if self.config.num_labels is None and self.config.id2label is None:
            raise ValueError("Both `num_labels` and `id2label` are None. Please provide at least one of them!")
        if self.config.id2label and self.config.num_labels is None:
            self.config.num_labels = len(self.config.id2label)
        bert_config = BertConfig(**self.config)
        return bert_config

    def forward(self, inputs, **kwargs) -> Dict:
        input_ids = inputs.get("token_ids")
        attention_mask = inputs.get("attention_mask", None)
        token_type_ids = inputs.get("token_type_ids", None)
        position_ids = inputs.get("position_ids", None)
        head_mask = inputs.get("head_mask", None)
        inputs_embeds = inputs.get("inputs_embeds", None)
        output_attentions = inputs.get("output_attentions", None)
        output_hidden_states = inputs.get("output_hidden_states", None)

        lm_outputs = self.bert(
            input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            **kwargs,
        )
        sequence_output = lm_outputs[0]

        sequence_output = self.dropout(sequence_output)
        logits = self.classifier(sequence_output)

        outputs = {
            "logits": logits,
            "hidden_states": lm_outputs.hidden_states,
            "attentions": lm_outputs.attentions,
            **inputs
        }
        return outputs

    def preprocess(self, inputs: Union[str, List[str]], **kwargs):
        if isinstance(inputs, str):
            inputs = [inputs]
        if "normalizer" in self.preprocessor:
            normalizer = self.preprocessor["normalizer"]
            inputs = normalizer(inputs)
        tokenizer = self.preprocessor["wordpiece_tokenizer"]
        inputs = tokenizer(
            inputs,
            return_word_ids=True,
            return_tokens=True,
            padding=True,
            truncation=True,
            return_tensors="pt",
            device=self.device,
        )
        return inputs

    def post_process(self, inputs, **kwargs):
        # TODO sequence labeling outputs should consider rejoining split words into single words with proper tag
        logits = inputs["logits"]
        tokens = inputs["tokens"]
        word_ids = inputs["word_ids"]  # noqa
        predictions = logits.argmax(2).cpu()
        predictions = [[self.config.id2label[p.item()] for p in prediction] for prediction in predictions]
        outputs = []
        for tokens_list, prediction in zip(tokens, predictions):
            results = []
            for token, tag in zip(tokens_list, prediction):
                if token not in self.config.prediction_skip_tokens:
                    results.append({"token": token, "tag": tag})
            outputs.append(results)
        return outputs



from dataclasses import dataclass
from typing import Iterable

from sklearn.metrics import recall_score

from ..configs import MetricConfig
from ..constants import MetricType
from ..registry import register_metric
from .metric import Metric


@dataclass
class RecallConfig(MetricConfig):
    name: str = MetricType.RECALL
    pos_label: int = None
    average: str = "macro"
    sample_weight: Iterable[float] = None


@register_metric("recall", config_class=RecallConfig)
class Recall(Metric):
    def __init__(self, config: RecallConfig, **kwargs):
        super().__init__(config, **kwargs)

    def compute(
        self,
        predictions=None,
        targets=None,
        labels=None,
        pos_label=None,
        average=None,
        sample_weight=None,
        zero_division="warn",
    ):

        pos_label = pos_label or self.config.pos_label
        average = average or self.config.average
        sample_weight = sample_weight or self.config.sample_weight

        score = recall_score(
            targets,
            predictions,
            labels=labels,
            pos_label=pos_label,
            average=average,
            sample_weight=sample_weight,
            zero_division=zero_division,
        )

        return float(score) if score.size == 1 else score



"""
A DistilBERT model for text classification built using HuggingFace Transformers
"""
from typing import Dict, List, Union

from torch import nn
from transformers import DistilBertConfig, DistilBertModel

from ....models import Model
from ....registry import register_model
from .distilbert_text_classification_config import DistilBertTextClassificationConfig


@register_model(model_name="distilbert_text_classification", config_class=DistilBertTextClassificationConfig)
class DistilBertTextClassification(Model):
    """
    A standard 🤗Transformers DistilBert model for text classification

    Args:
        config: The whole model config including arguments needed for the inner 🤗Transformers model.
    """

    def __init__(self, config: DistilBertTextClassificationConfig, **kwargs):
        super().__init__(config, **kwargs)
        self.distilbert = DistilBertModel(self._build_inner_config())
        self.pre_classifier = nn.Linear(self.config.dim, self.config.dim)
        self.classifier = nn.Linear(self.config.dim, self.config.num_labels)
        self.dropout = nn.Dropout(self.config.seq_classif_dropout)

    def _build_inner_config(self):
        if self.config.num_labels is None and self.config.id2label is None:
            raise ValueError("Both `num_labels` and `id2label` are None. Please provide at least one of them!")
        if self.config.id2label and self.config.num_labels is None:
            self.config.num_labels = len(self.config.id2label)
        bert_config = DistilBertConfig(**self.config)
        return bert_config

    def forward(self, inputs, **kwargs) -> Dict:
        input_ids = inputs.get("token_ids")
        attention_mask = inputs.get("attention_mask", None)
        head_mask = inputs.get("head_mask", None)
        inputs_embeds = inputs.get("inputs_embeds", None)
        output_attentions = inputs.get("output_attentions", None)
        output_hidden_states = inputs.get("output_hidden_states", None)

        lm_outputs = self.distilbert(
            input_ids,
            attention_mask=attention_mask,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            **kwargs,
        )
        hidden_state = lm_outputs[0]
        pooled_output = hidden_state[:, 0]
        pooled_output = self.pre_classifier(pooled_output)
        pooled_output = nn.ReLU()(pooled_output)
        pooled_output = self.dropout(pooled_output)
        logits = self.classifier(pooled_output)

        outputs = {
            "logits": logits,
            "hidden_states": lm_outputs.hidden_states,
            "attentions": lm_outputs.attentions,
        }
        return outputs

    def preprocess(self, inputs: Union[str, List[str]], **kwargs):
        if isinstance(inputs, str):
            inputs = [inputs]
        if "normalizer" in self.preprocessor:
            normalizer = self.preprocessor["normalizer"]
            inputs = normalizer(inputs)
        tokenizer = self.preprocessor["wordpiece_tokenizer"]
        inputs = tokenizer(inputs, return_tensors="pt", device=self.device)
        return inputs

    def post_process(self, inputs, **kwargs) -> Dict:
        logits = inputs["logits"]
        predictions = logits.argmax(1)
        predictions_probs = logits.softmax(1).max(1)
        outputs = {"labels": [], "probs": []}
        for prediction, prob in zip(predictions, predictions_probs):
            label = self.config.id2label[prediction.item()]
            outputs["labels"].append(label)
            outputs["probs"].append(prob.item())
        return outputs



import json
import os
from typing import Dict, Union

import omegaconf
from omegaconf import DictConfig

from ..constants import ConfigType, RegistryType
from .common_utils import snake_case
from .logging import get_logger


__all__ = [
    "flatten_dict",
    "load_yaml_config",
    "load_json_config",
    "hezar_config_to_hf_config",
    "get_module_config_class",
    "get_module_class",
    "get_registry_point",
]

logger = get_logger(__name__)


def flatten_dict(dict_config: Union[Dict, DictConfig]) -> DictConfig:
    """
    Flatten a nested Dict/DictConfig object

    Args:
        dict_config: A Dict/DictConfig object

    Returns:
        The flattened version of the dict-like object
    """

    config = DictConfig({})
    for k, v in dict_config.items():
        if isinstance(v, (Dict, DictConfig)):
            config.update(flatten_dict(v))
        else:
            config[k] = v

    return config


def load_yaml_config(path: Union[str, os.PathLike]):
    """
    Load yaml file using omegaconf
    """
    config = omegaconf.OmegaConf.load(path)
    return config


def load_json_config(path: Union[str, os.PathLike]):
    """
    Load json config file
    """
    with open(path) as f:
        config = json.load(f)
    return config


def hezar_config_to_hf_config(config):
    """
    Convert a :class:`hezar.Config` instance to HuggingFace :class:`transformers.PretrainedConfig` format

    Args:
        config: A :class:`Hezar.Config` instance
    """
    from transformers import PretrainedConfig

    hf_config = PretrainedConfig(**config)
    return hf_config


def get_registry_point(registry_key: str, registry_type: RegistryType):
    """
    Get the registry item by registry key name in a specific registry

    Args:
        registry_key: Module's name in the registry
        registry_type: Module's registry container type

    Returns:
        A Registry object
    """
    if registry_type == RegistryType.MODEL:
        from ..registry import models_registry  # noqa
        registry = models_registry

    elif registry_type == RegistryType.PREPROCESSOR:
        from ..registry import preprocessors_registry  # noqa
        registry = preprocessors_registry

    elif registry_type == RegistryType.DATASET:
        from ..registry import datasets_registry  # noqa
        registry = datasets_registry

    elif registry_type == RegistryType.EMBEDDING:
        from ..registry import embeddings_registry  # noqa
        registry = embeddings_registry

    elif registry_type == RegistryType.TRAINER:
        from ..registry import trainers_registry  # noqa
        registry = trainers_registry

    elif registry_type == RegistryType.METRIC:
        from ..registry import metrics_registry  # noqa
        registry = metrics_registry

    else:
        raise ValueError(f"Invalid `registry_type`: {registry_type}!")

    registry = registry[registry_key]
    return registry


def get_module_config_class(name: str, config_type: str = None):
    """
    Get the config class for a given module based on its registry name.

    Args:
        name (str): Module's registry name
        config_type (str): Registry type

    Returns:
        A class of type :class:`hezar.Config`
    """
    if config_type == ConfigType.MODEL:
        from ..registry import models_registry  # noqa
        registry = models_registry

    elif config_type == ConfigType.PREPROCESSOR:
        from ..registry import preprocessors_registry  # noqa
        registry = preprocessors_registry

    elif config_type == ConfigType.DATASET:
        from ..registry import datasets_registry  # noqa
        registry = datasets_registry

    elif config_type == ConfigType.EMBEDDING:
        from ..registry import embeddings_registry  # noqa
        registry = embeddings_registry

    elif config_type == ConfigType.METRIC:
        from ..registry import metrics_registry  # noqa
        registry = metrics_registry

    elif config_type == ConfigType.TRAINER:
        from ..registry import trainers_registry  # noqa
        registry = trainers_registry

    elif config_type is None:
        return None
    else:
        raise ValueError(f"Invalid `config_type`: {config_type}!")

    config_cls = registry[name].config_class
    return config_cls


def get_module_class(name: str, module_type: str):
    """
    Get module class based on registry name

    Args:
        name: Module's key name in its registry
        module_type: Type of the module e.g, model, dataset, preprocessor, embedding, etc

    Returns:
        A class corresponding to the given module
    """
    if module_type == ConfigType.MODEL:
        from ..registry import models_registry  # noqa
        registry = models_registry

    elif module_type == ConfigType.PREPROCESSOR:
        from ..registry import preprocessors_registry  # noqa
        registry = preprocessors_registry

    elif module_type == ConfigType.DATASET:
        from ..registry import datasets_registry  # noqa
        registry = datasets_registry

    elif module_type == ConfigType.EMBEDDING:
        from ..registry import embeddings_registry  # noqa
        registry = embeddings_registry

    elif module_type == ConfigType.METRIC:
        from ..registry import metrics_registry  # noqa
        registry = metrics_registry

    elif module_type == ConfigType.TRAINER:
        from ..registry import trainers_registry  # noqa
        registry = trainers_registry

    else:
        raise ValueError(f"Invalid `module_type`: {module_type}!")
    name = snake_case(name)
    module_cls = registry[name].module_class
    return module_cls



from dataclasses import dataclass
from typing import List, Optional, Union

from seqeval.metrics import accuracy_score, classification_report

from ..configs import MetricConfig
from ..constants import MetricType
from ..registry import register_metric
from ..utils import get_logger
from .metric import Metric


logger = get_logger(__name__)


@dataclass
class SeqevalConfig(MetricConfig):
    name: str = MetricType.SEQEVAL
    suffix: bool = False
    mode: Optional[str] = None
    sample_weight: Optional[List[int]] = None
    zero_division: Union[str, int] = 0
    digits: int = None


@register_metric("seqeval", config_class=SeqevalConfig)
class Seqeval(Metric):
    def __init__(self, config: SeqevalConfig, **kwargs):
        super().__init__(config, **kwargs)

    def compute(
        self,
        predictions=None,
        targets=None,
        suffix: bool = None,
        mode: Optional[str] = None,
        sample_weight: Optional[List[int]] = None,
        zero_division: Union[str, int] = None,
        digits: int = None,
        **kwargs,
    ):
        suffix = suffix or self.config.suffix
        mode = mode or self.config.mode
        sample_weight = sample_weight or self.config.sample_weight
        zero_division = zero_division or self.config.zero_division
        digits = digits or self.config.digits

        report = classification_report(
            y_true=targets,
            y_pred=predictions,
            suffix=suffix,
            output_dict=True,
            mode=mode,
            sample_weight=sample_weight,
            zero_division=zero_division,
        )
        report.pop("macro avg")
        report.pop("weighted avg")
        overall_score = report.pop("micro avg")

        results = {
            "accuracy": format(accuracy_score(predictions, targets)),
            "f1": overall_score["f1-score"],
            "recall": overall_score["recall"],
            "precision": overall_score["precision"],
        }
        if digits:
            results = {k: round(float(v), digits) for k, v in results.items()}
        return results



